program Hybrid_pef
  use sep

  !use solver_reg_nrm_bd_mod
  !use solver_reg_nrm_bd_prec_mod
  use solver_smp_nrm_mod
  use plane_search_mod
  use pef_bd_mod

  use t_conv_filter
  use util_axisIO

  implicit none

  integer :: n1,n2
  type(myaxis)  axis1, axis2

  ! <in.H Sparse_decon inv= filt=filter.H > out.H


  real, allocatable, dimension (:) :: xx,yy
  real, allocatable, dimension (:) :: filter, bb ! bb len = len(filter)-1
  real, allocatable, dimension (:) :: resd
  integer ::filter_len

  integer                          :: niter
  real                             :: qnt_d

  integer :: max_overshot_cnt
  !|r|, |d|
  !|F'C'(r), F'C'(d)|

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations

  logical ::inv
  character(12)  ::str_norm_dat
  integer        ::stat

  real, dimension(:), allocatable :: error

  type(myaxis) :: axis_wavelet, axis_filt_res
  integer  res_len
  real, dimension(:), allocatable :: filt_extend, wavelet,filt_res

  !Params
  call sep_init()

  call hinaxis(axis1,1)
  call hinaxis(axis2,2)
  n1 = axis1%n; n2 = axis2%n

  call from_param   ("inv",inv,.false.)

  if (inv) then
     call from_param ("niter", niter, 1); write(0,*) "niter=",niter
     call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

     call from_param ("qt_period",qt_period,1000); 
     write(0,*) "qt_period=",qt_period
     call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
     write(0,*) "max_overshot_cnt=", max_overshot_cnt

     call from_param   ("qnt_d", qnt_d, .9) 
     write(0,*) "quantile d = ", qnt_d

     call from_param   ("norm_dat", str_norm_dat, 'l2')
     write(0,*) "norm dat=", str_norm_dat

  end if

  call to_history ("label1","time(sec)")
  call to_history ("label2","x(km)")
  call to_history ("label3","Amp")


  !write(0,*) o1,d1,o2,d2,n1,n2
  call houtaxis(axis1,1)
  call houtaxis(axis2,2)

  allocate (xx(n1*n2), yy(n1*n2))
  xx=0; yy=0

  if (.not.inv) then !do the forward operator
     write(0,*) "Do forward modeling"
     !read in the reflectivity
     call sep_read (xx)

     !read in the filter
     call from_aux("filt", "n1", filter_len);
     allocate (filter(filter_len))
     call sep_read(filter,"filt");


     !Do the forward convolution
     call t_conv_filter_init(xx,n1,n2)
     stat = t_conv_filter_lop (.false.,.false.,filter,yy)

     call sep_write (yy)

     deallocate(filter)

  else !do the inversion with different scheme
     call from_param('filter_len',filter_len)

     allocate(bb(filter_len-1)) ! -1 since the leading 1 in filter is not counted
     allocate(filter(filter_len))
     allocate(resd(n1*n2))

     call sep_read(yy)

     !These terms only applies to solver_smp_nrm subroutine
     allocate (error (niter))
     write(0,*) "invoke the norm solver"


     if (exist_file("mod_iter")) then
        call houtaxis(axis1,1,'mod_iter')
        call houtaxis(axis2,2,'mod_iter')

        call to_history    ('o3',0,"mod_iter")
        call to_history    ('d3',1,"mod_iter")
        call to_history    ('n3',niter,"mod_iter")
        call to_history ("label3","Amplitude","mod_iter")
     end if

     call t_conv_filter_init(yy,n1,n2)

     bb=0

     write(0,*) "call the solver_smp_nrm"
     call solver_smp_nrm ( m=bb, d=-yy, Fop=t_conv_filter_pad0_lop, &
          & stepper=cgstep_ps, niter=niter ,&
          & norm_in=str_norm_dat , &
                                !& qnt_dat_in=1.0, qnt_mod_in=qnt_m, &
          & R_th_in = qnt_d,  & !Here we deal with the number directly 
          & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt,&
          & m0=bb, resd=resd) ;
     write(0,*) "Finish the call of solver_smp_nrm"

     !reassemble the xx and filter
     filter(1) = 1.0
     filter(2:filter_len) = bb(1:filter_len-1);


     call to_history('n1',filter_len,'filt')
     call sep_write(filter,'filt')
     call sep_write(resd)


     !Apply the forward operator to compute wavelet*filter
     if (exist_file('filt_out')) then
        if (exist_file('wavelet')) then
           call hinaxis(axis_wavelet,1,'wavelet')
           allocate(wavelet(axis_wavelet%n))
           call sep_read(wavelet,'wavelet')


           !We assume an arbitrary number of 128 input signal length
           !for simplicity
           res_len = 512
           allocate(filt_extend(res_len),filt_res(res_len))
           filt_extend = 0
           filt_extend(1:filter_len) = filter
           call t_conv_filter_init(filt_extend,res_len,1)
           stat = t_conv_filter_lop(.false.,.false.,wavelet,filt_res)

           !           write(0,*) 'filt_extend=', filt_extend
           !           write(0,*) 'wavelet=', wavelet
           !           write(0,*) 'result=', filt_res

           call set_axis(axis_filt_res,res_len,1.0,0.0,'sample')
           call houtaxis(axis_filt_res,1,'filt_out')
           call sep_write(filt_res,'filt_out')
           deallocate(wavelet,filt_extend,filt_res)
        else
           write(0,*) 'wavelet file does not\
           exist, can not produce ffilt_out!'
        end if
     end if

     deallocate(bb,filter,resd)

     !Get error Statistics
  end if

  deallocate (xx,yy)
  call sep_close()
end program Hybrid_pef
