# Kirchhoff -- Kirchhoff modeling and migration
#
# <in.H  Kirchhoff  adj=1 v0=1.5 [slowfit=slowfit.H] n3=1 halfdif=0 > out.H
#
#	adj=1	    	migration, i.e. data processing.
#
#	v0=1.5		velocity at surface.
#	alpha=0		vinterval(z) = v0 + alpha *z
#       slowfit.H =   	slowness(n1)  or  slowness(n1,n3)
#	halfdif=0	Don't use sqrt(d/dt) after forward modeling.
#	
# keyword Kirchhoff
# JFC
#%
integer n1, n2, n3, esize, pad2
from history:	integer n1, n2, n3=1, esize
if (esize !=4)	call erexit('esize != 4')
allocate:	real	data(n1,n2), modl(n1,n2), slow(n1), vrms(n1), vint(n1)
allocate:	complex datc(n1,n2), modc(n1,n2)
allocate:	real buff(n1)

subroutine dumb( n1,n2,n3, data, modl, slow, vrms, vint, buff)
integer	adj, i1,i2,i3, n1,n2,n3, halfdif
integer isvelfile, n2vel,auxpar
real	data(n1,n2), modl(n1,n2), o1,d1,o2,d2, v0, alpha, t
real	slow(n1), vrms(n1), vint(n1), buff(n1)
character*80 filename
from param:	integer adj=1, halfdif=0
from history:	real	o1,d1, o2,d2

isvelfile = getch('slowfit','s',filename)
if( isvelfile > 0 ) {
	i1=  auxpar('n2'   ,'i', n2vel, 'slowfit')
	call putch( 'n2vel','i', n2vel)
	if     ( n2vel == n3 )  call putlin('one velocity for each offset?')
	else if( n2vel == n2 )  call putlin('one velocity for each midpoint')
	else if( n2vel == 1  )  call putlin('one velocity for whole line')
	else                    call erexit('velocity_n2 \= data_n[23]')
	if(     n2vel == n2 ) 	call erexit('I cannot do v(x).')
	}
else {
	call putlin('Analytical v(z)')
	from param:	real 	v0=1.5, alpha=.5
	do i1= 1, n1  { t = o1+(i1-1)*d1 + d1/2.
		if( alpha*t != 0. )
		      slow(i1) = 1./( v0 * sqrt( (exp(alpha*t)-1.)/(alpha*t) ) )
		else
		      slow(i1) = 1./ v0
		}
	}

call hclose()

if( isvelfile>0  )
	if( n2vel != n3 )
		call sreed( 'slowfit', slow, 4*n1)

do i3= 1, n3 {					# Loop over offsets?
	if( isvelfile>0  ){
		if( n2vel == n3 )
			call sreed( 'slowfit', slow, 4*n1)}

	do i1= 1, n1
		vrms(i1) = 1./ amin1( 1./1.5, amax1(1./7.0, slow(i1)))

	call vint2rms(  1, 1.4, d1, vint, n1,  vrms )

	if( adj == 0)
			call sreed( 'in', modl, 4*n1*n2)
	else {
		if( halfdif > 0){
			do i2=1,n2 {
				call sreed( 'in',                    buff, 4*n1)
				call halfdifa( 1, 0, n1, data(1,i2), buff )
				}}

		else {
			call sreed( 'in', data, 4*n1*n2)
                }
	}	

	call kirchfast( adj, 0, vrms, o1,d1,d2, modl,n1,n2, data)

	if( adj == 0){
		if( halfdif > 0){
			do i2=1,n2 {
				call halfdifa( 0, 0, n1, data(1,i2), buff  )
				call srite('out',                    buff, 4*n1)
				}}
		else call srite('out', data, 4*n1*n2) 
	}
	else	call srite('out', modl, 4*n1*n2) 
}
return;	end

