!Pef program for Blind deconvolution

! The Problem is formulated as
!    [W*D   -W*I   ]  J [ a ]  - W*[ d ]  =  [ rd ]
!    [0     eps A  ]    [ m ]      [ 0 ]  =  [ rm ]

program Pef_bd
  use sep
  use helix
  use createhelixmod
  use bound
  use misinput
  use pef
  use compress
  use print

  use solver_reg_nrm_bd_mod
  use plane_search_mod
  use hconest
  use pef_bd_mod

  implicit none
  integer, dimension (:), allocatable :: n, n0, a, center, gap
  integer                             :: dim, n123, i, niter, na
  real,    dimension (:), pointer     :: dd, kk
  type( filter)                       :: aa   
  character (len = 6)                 :: varname
  real                                :: qnt_m,qnt_d

  integer :: len_pef
  real, dimension(:), pointer :: a_and_m, m
  real :: eps
  character(12) ::norm_d, norm_m
  character(128) :: buffer
  norm_d='hybrid'
  norm_m='hybrid'

  call sep_init (SOURCE)
  dim = sep_dimension ()
  allocate (n (dim), n0 (dim), center (dim), a (dim), gap (dim))
  call from_history (n)
  call to_history ("n",n)
  call from_par ("a", a)
  center = 1
  where (a (2:) > 1) center (:dim-1) = a(:dim-1)/2 + 1
  call from_par ("center", center, center)
  call from_par ("na", na, 0)
  if (na == 0) then
     write(0,*) 'call createhelix'
     gap = 0
     call from_par ("gap", gap, gap)
     aa = createhelix( n, center, gap, a) ! allocate PEF
     n0 = n
  else
     write(0,*) 'call allocatehelix'
     call allocatehelix (aa, na)
     call from_par ("lag", aa%lag)
     call from_par ("n", n0)
  end if

  n123 = product (n)
  write(0,*) "allocate dd,kk"
  allocate (dd (n123), kk (n123)) 
  call sep_read (dd)
  if (exist_file ("maskin")) then
     call sep_read (kk, "maskin")
  else
     kk = 1.
  end if
  write(0,*) "1"

  call boundn( n0, n, a, aa) 
  call find_mask( (kk /= 0.), aa)   ! missing data
  
  write(0,*)  "2"
  if (exist_file ("maskout")) then
     where (aa%mis) 
        kk = 0.
     elsewhere
        kk = 1.
     end where
     do i=1, dim
        call to_history (axisname(i),n(i),"maskout")
     end do
     call to_history ("esize",4,"maskout")
     call sep_write (kk, "maskout")
  end if
  
  write(0,*) 3
  call from_par( "niter", niter, size(aa%lag) )
  call from_param("qnt_d" , qnt_d  , 0.90)
  call from_param("qnt_m" , qnt_m  , 1.0 )
  call from_param("eps",  eps);
  write(0,*) "CHECK aa%lag and flt",aa%lag,aa%flt

  write(0,*) "The qnt_d/m input to PEF est is ", qnt_d," ",qnt_m,&
       & " niter=",niter," eps=",eps

  !Solve the inversion problem

  !For pef helicon, size(dd)==size(dd*pef=m) 
  len_pef = size(aa%lag)
  allocate(a_and_m( len_pef+size(dd))) 

  !Init the operators
  call hconest_init( dd, aa)

  write(0,*) "call the solver_reg_nrm_bd_mod"
  call solver_reg_nrm_bd( m=a_and_m, d=-dd, Fop=hconest_lop, &
       & Aop=unit_lop, stepper=cgstep_ps, nAop=n123, niter=niter, eps=eps,&
       & norm_dat_in=norm_d, norm_mod_in=norm_m, &
                                !& qnt_dat_in=, qnt_mod_in, &
       & R_th_d_in = qnt_d, R_th_m_in= qnt_m,  & !Here we deal with the number directly 
       & psiter=10, overshot_shrink_max_cnt_in=10,&
       & m0=a_and_m);
  write(0,*) "Finish the call of solver_reg_nrm_bd"

  !dissemble the vector m
  aa%flt = a_and_m(1:len_pef)
  !allocate m
  allocate(m(n123))  ! n123=total size of dd
  m = a_and_m(len_pef+1:)


  !Output the deconvolution result m
  !dim = sep_dimension ()
  !allocate (n (dim), n0 (dim), center (dim), a (dim), gap (dim))
  write(0,*) "output the deconv result"



  call to_history ("n1",n(1),"output_model");
  do i=2,dim
     write (buffer,"(a,i1)") 'n',i
     call to_history(buffer,n(i),"output_model");
  end do

  call sep_write(m,"output_model");

  deallocate(a_and_m, m)

  call resize( aa)                       ! eliminate zeroes
  call printn( n, center, a, aa)         ! print filter

  call to_history ("n1", size ( aa%lag))
  do i=2, dim
     write(0,*) "aixs name=", axisname(i)
     call to_history (axisname(i),1)
  end do
  call to_history ("lag", aa%lag)
  call sep_write (aa%flt)

  call sep_close ()
  deallocate (dd, kk)
  deallocate (n, n0, center, a, gap)
  call deallocatehelix( aa)
  call exit (0)
end program Pef_bd
