!Two step least-square optimization, update m and a sequentially

!Formulation, for every nonlinear iteration
!First Linearize the problem in A(M+deltaM) = d, invert for deltaM
!Then  linearize in alternate way M(A+deltaA) = d, invert for deltaA


program Sparse_decon_2_ls
  use sep
  use util_axisIO

  !use solver_reg_nrm_bd_mod
  use solver_reg_nrm_mod
  use plane_search_mod
  use mask1

  use t_conv_filter  !Operator M0
  use t_conv_data    !Operator A0

  use laplac1
  use op_sparse_decon_mod
  use pef_bd_mod

  implicit none

  integer :: n1,n2
  type(myaxis)  axis1, axis2, axis_iter

  ! <in.H Sparse_decon filt=filter.H > out.H


  real, allocatable, dimension (:) :: yy, Rd, Rma0, Rma, Rmm

  real, allocatable, dimension (:) :: a0,m0, delta_a, delta_m, data
  logical, pointer, dimension(:)   :: mask

  real, allocatable, dimension (:) :: resd, resm
  integer ::filter_len

  integer                          :: niter_nl,niter_lin, iter_nl

  real                             :: qnt_d,qnt_m
  real                             :: r_th_d, r_th_m, r_th_m_origi
  logical                          :: l_use_rthd, l_use_rthm


  integer :: zero_lag
  integer :: max_overshot_cnt
  !|r|, |d|
  !|F'C'(r), F'C'(d)|
  real, dimension(:),allocatable :: r_e,Fr_e,OF,OF_dat,OF_mod   ! recorded fitting error in every step, used for error plot
  real :: d_e,Fd_e,OF0,OF_dat0,OF_mod0

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations

  real    :: eps_a, eps_m, eps_m_origi
  real    :: alpha  ! scaling factor that accelerate convergence. [M0 A0/alpha] [ delta_a        ]
  ![ delta_m * alpha]
  character(12)  ::str_norm_mod, str_norm_dat, str_l2
  integer        ::stat

  real, dimension(:), allocatable :: error

  logical bsave
  integer i,j,k, nframe

  str_l2 = 'l2'

  !Params
  call sep_init()
  call hinaxis(axis1,1)
  call hinaxis(axis2,2)
  n1 = axis1%n; n2 = axis2%n

  call from_param   ("eps_a",   eps_a,    100.0)
  call from_param   ("eps_m",   eps_m_origi,    100.0)

  write(0,*) "eps a/m=", eps_a, eps_m_origi

  call from_param ("niter_nl", niter_nl, 1)
  call from_param ("niter_lin", niter_lin, 1)   !should be a small number
  write(0,*) "niter_nl/lin=",niter_nl, niter_lin

  call from_param ("psiter",psiter,2); write(0,*) "psiter=",psiter

  call from_param ("qt_period",qt_period,1000); 
  write(0,*) "qt_period=",qt_period
  call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
  write(0,*) "max_overshot_cnt=", max_overshot_cnt

  call from_param   ("qnt_d", qnt_d, .9) 
  call from_param   ("qnt_m", qnt_m, .9) 
  write(0,*) "quantile d/m = ", qnt_d,qnt_m

  call from_param ("r_th_d", r_th_d, -1.0)
  call from_param ("r_th_m", r_th_m_origi, -1.0)


  call from_param   ("norm_dat", str_norm_dat, 'l2')
  call from_param   ("norm_mod", str_norm_mod, 'hybrid')

  write(0,*) "norm dat/mod residue=", str_norm_dat, str_norm_mod

  call from_aux  ('wavelet_init', 'n1', filter_len)
  write(0,*) 'fiter len=',filter_len

  call from_param  ('zero_lag', zero_lag, 0)  !Default 0, means causal filter

  call houtaxis(axis1,1)
  call houtaxis(axis2,2)
  call houtaxis(axis1,1,'resd')
  call houtaxis(axis2,2,'resd')
  call houtaxis(axis1,1,'d_error')
  call houtaxis(axis2,2,'d_error')

  call to_history ("label1","time(sec)")
  call to_history ("label2","x(km)")
  call to_history ("label3","Amp")

  !SET THE DIMENSIONS of OUTPUT DUMP FILE
  nframe = (niter_nl-1)/10 + 2 !record at 0,1,11,21
  write(0,*) "nframe=",nframe
  call set_axis(axis_iter,nframe,1.0,1.0,"iter No.")

  call houtaxis(axis_iter,2,'wavelet')
  call houtaxis(axis_iter,3)
  call houtaxis(axis_iter,3,'resd')
  call houtaxis(axis_iter,3,'d_error') !Data fitting error

  call to_history('n1',filter_len,'wavelet')
  call to_history('d1',axis1%d,'wavelet')
  call to_history('o1',0.0,'wavelet')

  allocate(mask(filter_len))

  allocate (data(n1*n2), yy(n1*n2))
  allocate (m0(n1*n2), delta_m(n1*n2))

  allocate (resm(n1*n2),resd(n1*n2))
  allocate (Rd(n1*n2),Rmm(n1*n2))

  allocate (a0(filter_len),Rma0(filter_len),Rma(filter_len),delta_a(filter_len))

  data=0; yy=0; m0=0; delta_m=0; resd=0; resm=0; Rd=0; Rmm=0
  a0=0; delta_a=0; Rma0=0

  !read in the initial guess, filter, a0, reflectivity, m0
  !read in data
  if (exist_file('model_init')) then
     call sep_read (m0,'model_init')
  else
     m0 = 0
  endif

  call sep_read(a0,'wavelet_init');
  if (zero_lag .ne. 0) then
     a0 = 0
     a0(1+zero_lag)=1.0
  end if

  call sep_read(data)
  write(0,*) "max,min m0, a0, data"
  write(0,*) maxval(m0), maxval(a0), maxval(data)
  write(0,*) minval(m0), minval(a0), minval(data)

  do iter_nl=1, niter_nl+1
     !write(0,*) "invoke the norm solver"

     !!############################## Error dump
     if (exist_file("mod_iter")) then
        call houtaxis(axis1,1,'mod_iter')
        call houtaxis(axis2,2,'mod_iter')

        call to_history    ('o3',0,"mod_iter")
        call to_history    ('d3',1,"mod_iter")
        call to_history    ('n3', niter_nl*niter_lin ,"mod_iter")
        call to_history ("label3","Amplitude","mod_iter")
     end if
!!!!################################################## Error Dump

     !dump filter a0 and model m0 to file
     bsave = ((iter_nl.eq.1) .or. mod(iter_nl-2,10) .eq. 0)
     if (bsave) then
        call sep_write(a0,'wavelet')
        call sep_write(m0)
     end if

     !Dump resd from linearized inversion
     if (exist_file('resd') .and. bsave .and. (iter_nl.ne.1)) then
        call sep_write(resd,'resd');
     end if

     !Init the mask
     mask = .true.
     mask(1+zero_lag) = .false.
     call mask1_init(mask)

     l_use_rthd = .true.
     l_use_rthm = .true.
     if (abs( r_th_d - (-1)) < 1e-6) then 
        l_use_rthd = .false.
     endif
     if (abs( r_th_m_origi - (-1)) < 1e-6)  then
        l_use_rthm = .false.
     endif

     !INVERSION STEP I: Fix wavelet a, invert for delta_m
     !Do the forward modeling, yy = m0*a0
     call t_conv_filter_init(m0,n1,n2)
     stat = t_conv_filter_lop (.false.,.false.,a0,yy)  !yy is a0*m0
     Rd = data - yy
     write(0,*) 'max/min data,yy',maxval(data),maxval(yy),minval(data),minval(yy)
     write(0,*) 'max:min, resd', maxval(Rd), minval(Rd)

     !Dump resd at the first iter here     
     if (iter_nl.eq.1 .and. exist_file('resd')) then
        call sep_write(Rd(1:n1*n2),'resd')
     end if

     if (iter_nl .eq. niter_nl+1) then
        exit
     end if

     !Init opertor
     call t_conv_data_init(a0,n1,n2)

     !Generate Rm0
     stat = unit_lop(.false.,.false.,m0,Rmm)
     Rmm = eps_m_origi * Rmm

     r_th_m = r_th_m_origi * eps_m_origi
     delta_m = 0
     eps_m = eps_m_origi
     if ( l_use_rthd .and. l_use_rthm) then
        call solver_reg_nrm( m=delta_m, d=Rd, Fop=t_conv_data_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter_lin, eps=eps_m, rm0 = Rmm, &
             & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
                                !& qnt_dat_in=1.0, qnt_mod_in=qnt_m, &
             & R_th_d_in = r_th_d, R_th_m_in= r_th_m,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt, qt_period_in=qt_period,&
             & resd=resd)
     else if (l_use_rthd .and. (.not. l_use_rthm) ) then 
        call solver_reg_nrm( m=delta_m, d=Rd, Fop=t_conv_data_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter_lin, eps=eps_m, rm0 = Rmm,&
             & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_mod_in=qnt_m, &
             & R_th_d_in = r_th_d,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt, qt_period_in=qt_period,&
             & resd=resd)
     else if ((.not. l_use_rthd) .and.  l_use_rthm ) then
        call solver_reg_nrm( m=delta_m, d=Rd, Fop=t_conv_data_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter_lin, eps=eps_m, rm0 = Rmm,&
             & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_dat_in=qnt_d, &
             & R_th_m_in=r_th_m,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt, qt_period_in=qt_period,&
             & resd=resd)
     else   ! Use quantile both
        call solver_reg_nrm( m=delta_m, d=Rd, Fop=t_conv_data_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter_lin, eps=eps_m, rm0 = Rmm, &
             & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_dat_in=qnt_d, qnt_mod_in=qnt_m, &
             & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt, qt_period_in=qt_period,&
             & resd=resd)
     endif
     write(0,*) "Finish the call of solver_reg_nrm, step1"

     write(0,*) "max delta m", maxval(abs(delta_m))

     !update m0
     m0 = m0 + delta_m

     call solver_reg_nrm_close()

     !#####################
     !INVERSION 2, fix m, invert for wavelet a
     !Compute the data fitting error Rd first
     !Do the forward modeling, yy = m0*a0

     write(0,*) "enter the Inversion 2"
     call t_conv_filter_init(m0,n1,n2)
     stat = t_conv_filter_lop (.false.,.false.,a0,yy)  !yy is a0*m0
     Rd = data - yy

     write(0,*) "Apply the laplac regularization"
     !Apply the laplac1d to a0
     call laplac1_init(filter_len)
     Rma0 = 0
     stat = laplac1_lop(.false.,.false.,a0,Rma0)
     Rma = eps_a*Rma0 !The result of laplac

     !reset
     resd = 0; delta_a = 0

     write(0,*) "Call the solver in Inversion 2"
     !Call the solver to invert for delta_a
     call solver_reg_nrm( m=delta_a, d=Rd, Fop=t_conv_filter_lop, &
          & Aop=laplac1_lop, stepper=cgstep_ps, nAop=filter_len, niter=niter_lin/2, eps=eps_a, rm0 = Rma, Jop=mask1_lop, &
          & norm_dat_in=str_l2, norm_mod_in=str_l2, &
          & qnt_dat_in=qnt_d, &
          & psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt, qt_period_in=qt_period,&
          & resd=resd)
     call solver_reg_nrm_close()

     !Update a0
     a0 = a0 + delta_a

     !if (a0(1+zero_lag) .ne. 1.0) then
     !   write(0,*) "error, a0(1+zero_lag) !=1", a0(1+zero_lag)
     !end if

  end do ! non-linear iter loop

  deallocate(mask,data,yy,m0,delta_m,resm,Rd,resd, a0, delta_a, Rmm, Rma, Rma0)
  call sep_close()

end program Sparse_decon_2_ls
