program Sparse_decon
  use sep

  !use solver_reg_nrm_bd_mod
  use solver_reg_nrm_bd_prec_mod
  use plane_search_mod
  use pef_bd_mod

  use t_conv_filter
  use util_axisIO

  implicit none

  integer :: n1,n2
  type(myaxis)  axis1, axis2

  ! <in.H Sparse_decon inv= filt=filter.H > out.H


  real, allocatable, dimension (:) :: xx,yy, yy_shft
  real, allocatable, dimension (:) :: a_and_m  !filter combines xx
  real, allocatable, dimension (:) :: filter
  real, allocatable, dimension (:) :: d_conv_a, resd, resm
  integer ::filter_len

  integer                          :: niter

  real                             :: qnt_d,qnt_m
  real                             :: r_th_d, r_th_m
  logical                          :: l_use_rthd, l_use_rthm

  integer :: zero_lag
  integer :: max_overshot_cnt
  !|r|, |d|
  !|F'C'(r), F'C'(d)|
  real, dimension(:),allocatable :: r_e,Fr_e,OF,OF_dat,OF_mod   ! recorded fitting error in every step, used for error plot
  real :: d_e,Fd_e,OF0,OF_dat0,OF_mod0

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations

  real    :: eps
  real    :: alpha

  logical ::inv
  character(12)  ::str_norm_mod, str_norm_dat
  integer        ::stat

  real, dimension(:), allocatable :: error

  type(myaxis) :: axis_wavelet, axis_filt_res
  integer  res_len
  real, dimension(:), allocatable :: filt_extend, wavelet,filt_res

  integer i,j,k

  !Params
  call sep_init()

  call hinaxis(axis1,1)
  call hinaxis(axis2,2)
  n1 = axis1%n; n2 = axis2%n

  call from_param   ("eps",   eps,    100.0)
  write(0,*) "eps=", eps

  call from_param   ("inv",inv,.false.)

  if (inv) then
     call from_param ("niter", niter, 1); write(0,*) "niter=",niter
     call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

     call from_param ("qt_period",qt_period,5); 
     write(0,*) "qt_period=",qt_period
     call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
     write(0,*) "max_overshot_cnt=", max_overshot_cnt

     call from_param   ("qnt_d", qnt_d, .9) 
     call from_param   ("qnt_m", qnt_m, .9) 
     write(0,*) "quantile d/m = ", qnt_d,qnt_m

     call from_param ("r_th_d", r_th_d, -1.0)
     call from_param ("r_th_m", r_th_m, -1.0)

     call from_param   ("norm_dat", str_norm_dat, 'l2')
     call from_param   ("norm_mod", str_norm_mod, 'hybrid')
     call from_param   ('alpha', alpha)


     write(0,*) "norm dat/mod residue=", str_norm_dat, str_norm_mod
     write(0,*) "alpha=", alpha

  end if

  !write(0,*) o1,d1,o2,d2,n1,n2
  call houtaxis(axis1,1)
  call houtaxis(axis2,2)

  call to_history ("label1","time(sec)")
  call to_history ("label2","x(km)")
  call to_history ("label3","Amp")

  allocate (xx(n1*n2), yy(n1*n2), yy_shft(n1*n2))
  xx=0; yy=0

  if (.not.inv) then !do the forward operator
     write(0,*) "Do forward modeling"
     !read in the reflectivity
     call sep_read (xx)

     !read in the filter
     call from_aux("filt", "n1", filter_len);
     allocate (filter(filter_len))
     call sep_read(filter,"filt");


     !Do the forward convolution
     call t_conv_filter_init(xx,n1,n2)
     stat = t_conv_filter_lop (.false.,.false.,filter,yy)

     call sep_write (yy)

     deallocate(filter)

  else !do the inversion with different scheme
     call from_param  ('filter_len',filter_len)
     call from_param  ('zero_lag', zero_lag, 0)  !Default 0, means causal filter

     allocate(a_and_m(n1*n2+filter_len-1)) ! -1 since the leading 1 in filter is not counted
     allocate(filter(filter_len))
     allocate(resd(n1*n2),resm(n1*n2))

     call sep_read(yy)

     !These terms only applies to solver_smp_nrm subroutine
     allocate (error (niter))
     allocate (r_e(psiter*niter),OF(psiter*niter),OF_dat(psiter*niter),OF_mod(psiter*niter))
     allocate (Fr_e(niter))
     r_e = 0; Fr_e = 0; OF = 0; OF_dat = 0; OF_mod = 0
     d_e = 0; Fd_e = 0; OF0= 0; OF_dat0= 0; OF_mod0= 0
     write(0,*) "invoke the norm solver"


     if (exist_file("mod_iter")) then
        call houtaxis(axis1,1,'mod_iter')
        call houtaxis(axis2,2,'mod_iter')

        call to_history    ('o3',0,"mod_iter")
        call to_history    ('d3',1,"mod_iter")
        call to_history    ('n3',niter,"mod_iter")
        call to_history ("label3","Amplitude","mod_iter")
     end if

     call t_conv_filter_init(yy,n1,n2, zero_lag)

     yy_shft = 0
     
     yy_shft = yy
     if (zero_lag .eq. 0) then
        yy_shft = yy
     else !Need to shift yy along time lag
        do i=1,n2
           yy_shft((i-1)*n1+1+zero_lag:(i*n1)) = yy((i-1)*n1+1:i*n1-zero_lag)
        end do
     end if

     a_and_m=0

     l_use_rthd = .true.
     l_use_rthm = .true.
     if (abs( r_th_d - (-1)) < 1e-6) then 
        l_use_rthd = .false.
     endif
     if (abs( r_th_m - (-1)) < 1e-6)  then
        l_use_rthm = .false.
     endif


     r_th_m = r_th_m * eps

     !!Change back to quantile, whatever change here~~ ~~~~~~~~
     write(0,*) "call the solver_reg_nrm_bd_prec_mod"
     if ( l_use_rthd .and. l_use_rthm) then
        call solver_reg_nrm_bd_prec( m=a_and_m, d=-yy_shft, Fop=t_conv_filter_pad0_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
             & alpha=alpha, norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
                                !& qnt_dat_in=1.0, qnt_mod_in=qnt_m, &
             & R_th_d_in = r_th_d, R_th_m_in= r_th_m,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
             & m0=a_and_m, resd=resd, resm=resm)
     else if (l_use_rthd .and. (.not. l_use_rthm) ) then
        call solver_reg_nrm_bd_prec( m=a_and_m, d=-yy_shft, Fop=t_conv_filter_pad0_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
             & alpha=alpha, norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_mod_in=qnt_m, &
             & R_th_d_in = r_th_d,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
             & m0=a_and_m, resd=resd, resm=resm)
     else if ((.not. l_use_rthd) .and.  l_use_rthm ) then
        call solver_reg_nrm_bd_prec( m=a_and_m, d=-yy_shft, Fop=t_conv_filter_pad0_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
             & alpha=alpha, norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_dat_in= qnt_d,  &
             & R_th_m_in= r_th_m,  & !Here we deal with the number directly 
             & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
             & m0=a_and_m, resd=resd, resm=resm)
     else   ! Use quantile both
        call solver_reg_nrm_bd_prec( m=a_and_m, d=-yy_shft, Fop=t_conv_filter_pad0_lop, &
             & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
             & alpha=alpha, norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
             & qnt_dat_in=qnt_d, qnt_mod_in=qnt_m, &
             & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
             & m0=a_and_m, resd=resd, resm=resm)
     endif
     write(0,*) "Finish the call of solver_reg_nrm_bd"

     !reassemble the xx and filter
     if (zero_lag > 0) then
        filter(1:zero_lag) = a_and_m(1:zero_lag)
     end if
     filter(1+zero_lag) = 1.0
     if (zero_lag < filter_len-1) then
        filter(zero_lag+2:(filter_len)) = a_and_m(zero_lag+1:filter_len-1)
     end if

     xx = a_and_m(filter_len:);

     call to_history('n1',filter_len,'filt')
     call sep_write(filter,'filt')
     call sep_write(xx)
     
     !Apply the forward operator to compute d*a
     if (exist_file('d_a')) then
        call houtaxis(axis1,1,'d_a')
        call houtaxis(axis2,2,'d_a')
        allocate(d_conv_a(n1*n2))
        call t_conv_filter_init(yy,n1,n2)
        stat = t_conv_filter_lop(.false.,.false.,filter,d_conv_a)

        call sep_write(d_conv_a,'d_a')
        deallocate(d_conv_a)
     end if

     !Apply the forward operator to compute wavelet*filter
     if (exist_file('filt_out')) then
        if (exist_file('wavelet')) then
           call hinaxis(axis_wavelet,1,'wavelet')
           allocate(wavelet(axis_wavelet%n))
           call sep_read(wavelet,'wavelet')


           !We assume an arbitrary number of 512 input signal length
           !for simplicity
           res_len = 512
           allocate(filt_extend(res_len),filt_res(res_len))
           filt_extend = 0
           filt_extend(1:filter_len) = filter
           call t_conv_filter_init(filt_extend,res_len,1)
           stat = t_conv_filter_lop(.false.,.false.,wavelet,filt_res)

           !           write(0,*) 'filt_extend=', filt_extend
           !           write(0,*) 'wavelet=', wavelet
           !           write(0,*) 'result=', filt_res

           call set_axis(axis_filt_res,res_len,1.0,0.0,'sample')
           call houtaxis(axis_filt_res,1,'filt_out')
           call sep_write(filt_res,'filt_out')
           deallocate(wavelet,filt_extend,filt_res)
        else
           write(0,*) 'wavelet file does not\
           exist, can not produce ffilt_out!'
        end if
     end if
      write(0,*) '12'
     !Dump resd, resm
     if (exist_file('resd')) then
        call houtaxis(axis1,1,'resd')
        call houtaxis(axis2,2,'resd')
        call sep_write(resd,'resd');
     end if

     if (exist_file('resm')) then
        call houtaxis(axis1,1,'resm')
        call houtaxis(axis2,2,'resm')
        call sep_write(resm,'resm');
     end if

     deallocate(a_and_m,filter, resd, resm)




     !Get error Statistics
     stat = reg_nrm_get_error_stat(OF0,OF,OF_dat0,OF_dat,OF_mod0,OF_mod,d_e,r_e,Fd_e,Fr_e)

     call solver_reg_nrm_bd_prec_close()


     !Compute 1-|r|/|d|
     if (exist_file("avg_r")) then
        write(0,*) "output the avg_r, avg_r0=",d_e
        call to_history("n1",psiter*niter,"avg_r")
        call to_history("o1",1,"avg_r")
        call to_history("d1",1,"avg_r")
        call to_history("avg_r0",d_e,"avg_r")
        allocate(tmp_ar(niter*psiter))
        tmp_ar = 1 - r_e/d_e
        write(0,*) "r_e=",r_e, "d_e=",d_e
        call sep_write(tmp_ar,"avg_r")
        deallocate(tmp_ar)
     end if

     !Compute 1-|F'C'(r)|/|F'C'(d)|
     if (exist_file("meas_g")) then
        write(0,*) "output the meas_g,meas_g0=",Fd_e

        call to_history("n1",niter,"meas_g")
        call to_history("o1",1,"meas_g")
        call to_history("d1",1,"meas_g")
        call to_history("meas_g0",Fd_e,"meas_g")
        allocate(tmp_ar(niter))
        tmp_ar = 1 - Fr_e/Fd_e
        write(0,*) "Fr_e=",Fr_e, "Fd_e=",Fd_e
        call sep_write(tmp_ar,"meas_g")
        deallocate(tmp_ar)
     end if

     !Output the Objective Function value
     if (exist_file("of")) then
        write(0,*) "output the objective function value=", OF0, OF
        call to_history("n1",niter*psiter,"of")
        call to_history("o1",1,"of")
        call to_history("d1",1,"of")
        call to_history("n2",3,'of')

        call to_history("objective_func_total_0",OF0,"of")
        call to_history("objective_func_dat_0",OF_dat0,"of")
        call to_history("objective_func_dat_0",OF_mod0,"of")
        call sep_write(OF,"of")
        call sep_write(OF_dat,"of")
        call sep_write(OF_mod,"of")
     end if

     deallocate (r_e,OF,OF_dat,OF_mod)
     deallocate (Fr_e)

  end if
  
  deallocate (xx,yy,yy_shft)
  call sep_close()
end program Sparse_decon
