!Given the src signature, invert the refl model only
program Sparse_decon_refl
  use sep

  !use solver_reg_nrm_bd_mod
  use solver_reg_nrm_mod
  use plane_search_mod
  use pef_bd_mod

  use t_conv_data
  use util_axisIO

  implicit none

  integer :: n1,n2
  type(myaxis)  axis1, axis2

  ! <in.H Sparse_decon inv= filt=filter.H > out.H


  real, allocatable, dimension (:) :: xx,yy
  real, allocatable, dimension (:) :: filter
  real, allocatable, dimension (:) :: resd, resm
  integer ::filter_len

  integer                          :: niter

  real                             :: qnt_d,qnt_m
  real                             :: r_th_d, r_th_m
  logical                          :: l_use_rthd, l_use_rthm

  integer :: max_overshot_cnt

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations

  real    :: eps

  logical ::inv
  character(12)  ::str_norm_mod, str_norm_dat
  integer        ::stat

  real, dimension(:), allocatable :: error

  integer i,j,k

  !Params
  call sep_init()

  call hinaxis(axis1,1)
  call hinaxis(axis2,2)
  n1 = axis1%n; n2 = axis2%n

  call from_param   ("eps",   eps,    100.0)
  write(0,*) "eps=", eps

  call from_param ("niter", niter, 1); write(0,*) "niter=",niter
  call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

  call from_param ("qt_period",qt_period,1000); 
  write(0,*) "qt_period=",qt_period
  call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
  write(0,*) "max_overshot_cnt=", max_overshot_cnt

  call from_param   ("qnt_d", qnt_d, .9) 
  call from_param   ("qnt_m", qnt_m, .9) 
  write(0,*) "quantile d/m = ", qnt_d,qnt_m

  call from_param ("r_th_d", r_th_d, -1.0)
  call from_param ("r_th_m", r_th_m, -1.0)

  call from_param   ("norm_dat", str_norm_dat, 'l2')
  call from_param   ("norm_mod", str_norm_mod, 'hybrid')

  write(0,*) "norm dat/mod residue=", str_norm_dat, str_norm_mod

  !write(0,*) o1,d1,o2,d2,n1,n2
  call houtaxis(axis1,1)
  call houtaxis(axis2,2)

  call to_history ("label1","time(sec)")
  call to_history ("label2","x(km)")
  call to_history ("label3","Amp")

  allocate (xx(n1*n2), yy(n1*n2))
  xx=0; yy=0

  call from_aux('filt','n1',filter_len)
  allocate(filter(filter_len))
  allocate(resd(n1*n2),resm(n1*n2))

  call sep_read(yy)
  call sep_read(filter,'filt')


  if (exist_file("mod_iter")) then
     call houtaxis(axis1,1,'mod_iter')
     call houtaxis(axis2,2,'mod_iter')

     call to_history    ('o3',0,"mod_iter")
     call to_history    ('d3',1,"mod_iter")
     call to_history    ('n3',niter,"mod_iter")
     call to_history ("label3","Amplitude","mod_iter")
  end if

  call t_conv_data_init(filter,n1,n2)

  l_use_rthd = .true.
  l_use_rthm = .true.
  if (abs( r_th_d - (-1)) < 1e-6) then 
     l_use_rthd = .false.
  endif
  if (abs( r_th_m - (-1)) < 1e-6)  then
     l_use_rthm = .false.
  endif


  r_th_m = r_th_m * eps

  !!Change back to quantile, whatever change here~~ ~~~~~~~~
  write(0,*) "call the solver_reg_nrm"
  if ( l_use_rthd .and. l_use_rthm) then
     call solver_reg_nrm( m=xx, d=yy, Fop=t_conv_data_lop, &
          & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
          & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
                                !& qnt_dat_in=1.0, qnt_mod_in=qnt_m, &
          & R_th_d_in = r_th_d, R_th_m_in= r_th_m,  & !Here we deal with the number directly 
          & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
          & m0=xx, resd=resd, resm=resm)
  else if (l_use_rthd .and. (.not. l_use_rthm) ) then
     call solver_reg_nrm( m=xx, d=yy, Fop=t_conv_data_lop, &
          & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
          & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
          & qnt_mod_in=qnt_m, &
          & R_th_d_in = r_th_d,  & !Here we deal with the number directly 
          & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
          & m0=xx, resd=resd, resm=resm)
  else if ((.not. l_use_rthd) .and.  l_use_rthm ) then
     call solver_reg_nrm( m=xx, d=yy, Fop=t_conv_data_lop, &
          & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
          & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
          & qnt_dat_in= qnt_d,  &
          & R_th_m_in= r_th_m,  & !Here we deal with the number directly 
          & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
          & m0=xx, resd=resd, resm=resm)
  else   ! Use quantile both
     call solver_reg_nrm( m=xx, d=yy, Fop=t_conv_data_lop, &
          & Aop=unit_lop, stepper=cgstep_ps, nAop=n1*n2, niter=niter, eps=eps,&
          & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, &
          & qnt_dat_in=qnt_d, qnt_mod_in=qnt_m, &
          & psiter=psiter, overshot_shrink_max_cnt_in=8, qt_period_in=qt_period,&
          & m0=xx, resd=resd, resm=resm)
  endif
  write(0,*) "Finish the call of solver_reg_nrm_bd"
  call sep_write(xx)

  !Dump resd, resm
  if (exist_file('resd')) then
     call houtaxis(axis1,1,'resd')
     call houtaxis(axis2,2,'resd')
     call sep_write(resd,'resd');
  end if

  if (exist_file('resm')) then
     call houtaxis(axis1,1,'resm')
     call houtaxis(axis2,2,'resm')
     call sep_write(resm,'resm');
  end if

  deallocate(filter, resd, resm)

  call solver_reg_nrm_close()

  deallocate (xx,yy)
  call sep_close()
end program Sparse_decon_refl
