program main
  use base_struct_mod
  use base_weight_mod
  use base_mskopr_mod
  use hes_convol_mod
  use hes_reglar_mod
  use solver_reg_mod
  use cgstep_mod
  use sep
  implicit none

  type(hes_grid)                    :: hgrid
  real, dimension(:)  , allocatable :: modl  !!.. model space, the reflectivity model
  real, dimension(:)  , allocatable :: data  !!.. data space , the migrated image
  real, dimension(:)  , allocatable :: mask  !!.. mask
  real, dimension(:,:), allocatable :: mmov  !!.. model movie
  real, dimension(:,:), allocatable :: rmov  !!.. residual movie
  logical                           :: adj, search, vareps
  character(len=256)                :: modl_, data_, hess_, mask_
  character(len=256)                :: mmov_, rmov_ 
  integer                           :: stat, ndim, iter, niter, reglar
  real                              :: imgsign, eps 
  double precision                  :: t1, t2

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()
  call initpar()

  modl_ = "modl" 
  data_ = "data"
  hess_ = "hess"
  mask_ = 'mask'
  mmov_ = 'mmov'
  rmov_ = 'rmov'
 
  call from_param("adj"     , adj     , .true.)
  write(0,*) "adj=",adj

  call from_param("imgsign" , imgsign ,    +1.)
  call from_param("niter"   , niter   ,    50 )
  call from_param("reglar"  , reglar  ,      1)
  call from_param("eps"     ,      eps,  0.001)
  
  call auxinout(modl_)
  call auxinout(data_)
  call auxin   (hess_)
  call auxin   (mask_)
  call auxout  (mmov_)
  call auxout  (rmov_)
 

  call hinaxis(hess_, hgrid%ax_modl, 1) 
  call hinaxis(hess_, hgrid%ay_modl, 2)
  call hinaxis(hess_, hgrid%az_modl, 3)
  call hinaxis(hess_, hgrid%ax_offd, 4)
  call hinaxis(hess_, hgrid%ay_offd, 5)
  call hinaxis(hess_, hgrid%az_offd, 6)

  ndim  = hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n

  allocate(modl(ndim))
  allocate(data(ndim))
  allocate(mask(ndim))
  allocate(mmov(ndim,niter))
  allocate(rmov(ndim,niter))

  modl = 0.; data = 0.; mask = 0.
  mmov = 0.; rmov = 0.

  if (adj) then
     call houaxis(modl_, hgrid%ax_modl, 1)
     call houaxis(modl_, hgrid%ay_modl, 2)
     call houaxis(modl_, hgrid%az_modl, 3)

     call houaxis(mmov_, hgrid%ax_modl, 1)
     call houaxis(mmov_, hgrid%ay_modl, 2)
     call houaxis(mmov_, hgrid%az_modl, 3)
     call auxputch("n4", "i", niter, mmov_)

     call houaxis(rmov_, hgrid%ax_modl, 1)
     call houaxis(rmov_, hgrid%ay_modl, 2)
     call houaxis(rmov_, hgrid%az_modl, 3)
     call auxputch("n4", "i", niter, rmov_)

     call sep_read(data, data_) 
     !The data should be the migrated Image L'd
     data = data * imgsign / maxval(abs(data))  !The data is normalized here, why normalize ??

  else
     call houaxis(data_, hgrid%ax_modl, 1)
     call houaxis(data_, hgrid%ay_modl, 2)
     call houaxis(data_, hgrid%az_modl, 3)
     call sep_read(modl, modl_)
  end if

  stat = hes_convol_init(hgrid, hess_)

  if (adj) then
     write(0,*) 'Doing Hessian Inversion'
     !Doing inversion
     call sep_read(mask, mask_)
     !!.. mask operator which masks out the salt body
     stat = base_mskopr_init(mask)
     !!.. regularization operator, reglar = 1 for damping, 2 for horizontal laplacian
     stat = hes_reglar_init(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n,reglar)

     !!!----INTERFACE-----------
     !solver_reg( m,d, Fop, Aop, stepper, nAop, niter,eps   ,  &
     !&           Wop,Jop,m0,rm0,err,resd,resm,mmov,rmov,verb)

     call solver_reg( m=modl,d=data, Fop=hes_convol_oper, Aop=hes_reglar_oper, &
                    & stepper=cgstep, nAop=ndim, niter=niter, eps=eps, &
                    & WOp=base_mskopr_oper, mmov=mmov, rmov=rmov, verb=.true.)
!!--------------------------------------------------------------------------------------
     stat = base_mskopr_clos()

  else
     !Doing forward modeling
     write(0,*) 'doing forward modeling'
     stat = hes_convol_oper(adj,.false.,modl,data)
  end if

  stat = hes_convol_clos()

  if (adj) then
     call sep_write(modl, modl_)
     call sep_write(mmov, mmov_)
     call sep_write(rmov, rmov_)
  else
     call sep_write(data, data_)
  end if

  deallocate(modl, data)

end program
