module laplac1
! Laplacian operator in 1-D
  use adj_mod
  use igrad1
  implicit none
  logical,   parameter, private   :: T = .true. 
  logical,   parameter, private   :: F = .false. 
  real ,private,  dimension (:), allocatable  :: tmp 
  contains
  subroutine laplac1_init ( m1)
    integer m1
    if (.not. allocated(tmp)) then
      allocate(tmp(m1))
    end if
  end subroutine

  function laplac1_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplac1_lop2(adj,add,x,y )
    stat=0
  end function 
  subroutine laplac1_lop2(adj,add,x,y)
    logical,intent(in) :: adj,add 
    real, dimension (:)  :: x 
    real, dimension (:)  :: y 
    integer stat1, stat2
    if ( adj) then
      stat1 = igrad1_lop (  F,    F, y, tmp)   ! tmp = grad y
      stat2 = igrad1_lop (  T,  add, x, tmp)   ! x = x + grad' tmp
    else
      stat1 = igrad1_lop (  F,    F, x, tmp)   ! tmp = grad x
      stat2 = igrad1_lop (  T,  add, y, tmp)   ! y = y + grad' tmp
    end if
  end subroutine 
  subroutine laplac1_close()
    deallocate( tmp )
  end subroutine 
end module 
