module op_sparse_decon_mod
  
  use adj_mod
  use pef_bd_mod
  use t_conv_filter
  
  use t_conv_data
  use t_conv2_mod

  use laplac1

  use chain_mod_1

  implicit none
  integer, private :: n1,n2, filter_len,n_Rma !(length of the regularization vector on filter a
  real, private :: eps_a !, alpha
  integer, private :: reg_type !(regularization for the wavelet)

contains

  subroutine op_MA_init( n1_in, n2_in, filter_len_in, eps_a_in, n_Rma_in, reg_in)
    integer :: n1_in, n2_in, filter_len_in, n_Rma_in
    real :: eps_a_in
    integer:: reg_in

    n1 = n1_in
    n2 = n2_in
    filter_len = filter_len_in

    n_Rma = n_Rma_in
    eps_a = eps_a_in

    reg_type = reg_in

  end subroutine op_MA_init

  

  function op_MA( adj, add, xx, yy) result(stat)  ! [M and A]
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy
    !xx should be at the length of a_and_m
    call adjnull(adj,add,xx,yy)
    call row0( t_conv_filter_lop, t_conv_data_lop, adj, add, xx(1:filter_len), xx(filter_len+1:), yy, 1.0)  !eps=1.0
    !write(0,*) "sizes", size(xx), size(xx(1:filter_len)), size(xx(1+filter_len:))
    stat = 0
  end function op_MA


!Function Interface for regularzation wavelet
  function wavelet_reg(adj,add,xx,yy) result(stat) 
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy

    select case (reg_type)
    case (1)  !damping regularization
       stat = unit_lop(adj,add,xx,yy)
    case (2)  !Spectrum data regularization
       stat = t_conv2_lop(adj,add,xx,yy)
    case (3)  ! 1D laplacian
       stat = laplac1_lop(adj,add,xx,yy)
    case default
       write(0,*) '!!! reg_type param error'
    end select

  end function wavelet_reg
  

  !Here add the regularization for the wavelet
  function op_MA_and_Rma(adj,add,xx,yy) result(stat) !here yy=length(Rma)+len(rd)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy
    integer ::st
    integer ::len_rd

    call adjnull(adj,add,xx,yy)
    len_rd = n1*n2
    if (size(yy) .ne. (n_Rma+len_rd) ) then 
       write(0,*) "size(yy),n_Rma,len_rd", size(yy), n_Rma, len_rd
       call erexit("size(yy) is not equal to n_Rma+n_Rd, sth is wrong")
    end if

    
    if (adj) then
       st = op_MA(.true., add, xx, yy(1:len_rd))
       st = wavelet_reg(.true.,.true.,xx(1:filter_len), eps_a*yy(len_rd+1:))
    else
       st = op_MA(.false., add, xx, yy(1:len_rd))
       st = wavelet_reg(.false.,add,eps_a*xx(1:filter_len),yy(len_rd+1:))
    end if
    stat = 0
  end function op_MA_and_Rma

!!!! The regulraization on delta_m, just a unit operator on model m
  function reg_delta_m_lop( adj, add, xx, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy
    !xx is with length a_and_m, yy is with length m
    call adjnull(adj,add,xx,yy)
    if (adj) then
       xx(filter_len+1:) = xx(filter_len+1:) + yy
    else
       yy = yy + xx(filter_len+1:)
    end if
    stat=0
  end function reg_delta_m_lop
end module op_sparse_decon_mod
