!The regularized version of the hybrid norm solver

module solver_reg_nrm_mod 
  ! 0 = W (F J m - d)
  ! 0 =    A   m

  use sep
  use chain0_mod 
  use solver_report_mod
  use norm_mod
  use plane_search_mod

  implicit none
  logical, parameter, private  :: T = .true., F = .false.


  !store error information, currently avg_r_dat, avg_m_dat are not computed
  real,private		::avg_rd0, avg_rm0, meas_g0,sum_norm_dat0, sum_norm_mod0, sum_norm0 ! |r| and |F'C'(r)| intial value
  real,dimension(:),allocatable,target,private :: avg_rd, avg_rm, sum_norm_dat, sum_norm_mod, sum_norm
  real,dimension(:),allocatable,private :: meas_g ! the norm of F'C'(r) ,using L2 currently
  
  contains
    subroutine solver_reg_nrm( m,d, Fop, Aop, stepper, nAop, niter,eps   ,  &
         &           norm_dat_in,norm_mod_in,qnt_dat_in,qnt_mod_in,R_th_d_in,R_th_m_in, qt_period_in,psiter,overshot_shrink_max_cnt_in, &
         &           Wop,Jop,m0,rm0,err,resd,resm,mmov,rmov,verb)
      optional ::    norm_dat_in,norm_mod_in,qnt_dat_in,qnt_mod_in,R_th_d_in,R_th_m_in, qt_period_in,psiter,overshot_shrink_max_cnt_in, &
           &         Wop,Jop,m0,rm0,err,resd,resm,mmov,rmov,verb

      interface  

         !-------------------- begin definitions --------------------
         integer function Fop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Fop
         integer function Aop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Aop
         integer function Wop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Wop
         integer function Jop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Jop
         integer function stepper(forget,m,g,rr,gg)
           real, dimension(:)  ::        m,g,rr,gg    
           logical             :: forget
         end function stepper
      end interface
      real, dimension(:),    intent(in)          :: d, m0,rm0

      ! Iter. for solver and plane search    
      integer,               intent(in)          :: niter, nAop, psiter  
      logical,               intent(in)          :: verb
      real,                  intent(in)          :: eps

      character(12)                              :: norm_dat_in     ! type of Norm to be used
      character(12)                              :: norm_mod_in     ! type of Norm to be used
      character(12)                              :: str_norm_dat     ! type of Norm to be use
      character(12)                              :: str_norm_mod     ! type of Norm to be use
      character(12)  ::str_l2

      real,                  intent(in)          :: qnt_dat_in         ! Quantile to set R_thr (R_bar)
      real,                  intent(in)          :: qnt_mod_in         ! Quantile to set R_thr (R_bar)
      real :: r_th_d_in, r_th_m_in

      integer,               intent(in)          :: overshot_shrink_max_cnt_in

      integer ::  qt_period_in
      integer ::  qt_period
      real, dimension(:),    intent(out)         :: m,err, resd,resm
      real, dimension(:,:),  intent(out)         ::        rmov,mmov
      real, dimension(size(d))                   :: rwd
      real, dimension(nAop)                      :: rwm
      real, dimension(size( m))                  :: g
      real, dimension(size( d) + nAop), target   :: rr, gg, tt
      real, dimension(:), pointer                :: rd, gd, td
      real, dimension(:), pointer                :: rm, gm, tm

      real, dimension(:), pointer                :: g_tmp
      integer                                    :: iter, stat, psiter0
      logical                                    :: forget

      real                                       :: rt_d,rt1_d,rt2_d,quant_d  ! R_thr (R_bar), Quantile
      real                                       :: rt_m,rt1_m,rt2_m,quant_m  ! R_thr (R_bar), Quantile


      type(norm_obj) :: rd_norm   !norm object for rd
      type(norm_obj) :: rm_norm   !norm object for model residue
      type(norm_obj) :: g_norm    !L2 norm for computing the gradient, measuring success
      real           :: R_tmp
      real, dimension(:), pointer ::ptr,ptr1,ptr2,ptr3,ptr4

      logical ::L_need_Rth_m, L_need_Rth_d
      logical ::L_need_find_qnt_m,L_need_find_qnt_d

      !Assign the array pointers
      rd => rr(1:size(d))
      rm => rr(1+size(d):)
      gd => gg(1:size(d))
      gm => gg(1+size(d):)
      td => tt(1:size(d))
      tm => tt(1+size(d):)

      ! -------- begin initialization --------
      psiter0=1
      if (present(psiter)) psiter0=psiter

      if (present(Wop)) then
         write(0,*) "Wop present, before Wop: value rd is ", d(size(d))
         stat=Wop(F,F,-d,rd) 
         write(0,*) "After Wop: value of rd is ", rd(size(d))
      else
         rd = -d                              !Rd  = -W  d
      end if
      rm = 0.
      if (present(rm0)) then
         rm=rm0                               !Rm  =    Rm0
      end if
      if (present( m0)) then
         m=m0                                 !m   =     m0
         if (present(Wop)) then
            call chain0(Wop,Fop,F,T,m,rd,td)
         else
            stat=           Fop(F,T,m,rd   )   !Rd +=  WF m0
         end if
         stat = Aop(F,T,eps*m0,rm)                           
         !Rm += e A m0
      else
         m=0
      end if
      forget = T


      ! --------------- Set Norm quantile and rt -------------
      if (.not. present(norm_dat_in)) then
         str_norm_dat="hybrid"
      else
         str_norm_dat = norm_dat_in
      end if
      if (.not. present(norm_mod_in)) then
         str_norm_mod ="hybrid"
      else
         str_norm_mod = norm_mod_in
      end if

      quant_m=1.00
      quant_d=1.00
      rt_m = 1.0; rt_d = 1.0
      L_need_find_qnt_d = .false.; L_need_find_qnt_m = .false.
      L_need_Rth_d = norm_need_Rth(str_norm_dat)
      L_need_Rth_m = norm_need_Rth(str_norm_mod)

      if (L_need_Rth_d) then
         if (present(R_th_d_in)) then
            rt_d = R_th_d_in
            L_need_find_qnt_d = .false.
         else
            L_need_find_qnt_d = .true.
            if (present(qnt_dat_in)) then
               quant_d = qnt_dat_in
            endif
         end if
      end if

      if (L_need_Rth_m) then
         if (present(R_th_m_in)) then
            rt_m = R_th_m_in
            L_need_find_qnt_m = .false.
         else
            L_need_find_qnt_m = .true.
            if (present(qnt_mod_in)) then
               quant_m = qnt_mod_in
            endif
         end if
      end if

      if (.not.(quant_m <= 1.0 .and. quant_m >= 0)) then
         call erexit("!!!!quant_m value is invalid, check the param")
      end if
      if (.not.(quant_d <= 1.0 .and. quant_d >= 0)) then
         call erexit("!!!!quant_d value is invalid, check the param")
      end if

      !Initalization error recording array
      allocate(avg_rd(psiter0*niter),avg_rm(psiter0*niter),sum_norm_dat(psiter0*niter), sum_norm_mod(psiter0*niter),sum_norm(psiter0*niter));
      allocate(meas_g(niter))

      write(0,*) "init g_norm l2 norm"
      str_l2 = "l2"
      call norm_init(g_norm,0.0,str_l2)

      qt_period = 10000 !Default don't search Rth for multiple times
      if (.not. present(qt_period_in)) then
         write(0,*) "qt_change_period not specified"
      else
         qt_period = qt_period_in
      end if

      !Init the plane_search_module, interface is as below:
      !ps_init(psiter_in, ps_shrinkthres_in, len_datres_in, len_modres_in, len_mod, norm_datres_in, norm_modres_in)
      if (.not.present(overshot_shrink_max_cnt_in)) then
         call ps_init(psiter0,0,size(rd),  size(rm),             size(m),rd_norm,        rm_norm)
      else 
         call ps_init(psiter0,overshot_shrink_max_cnt_in,size(rd),size(rm),size(m), rd_norm,        rm_norm)
      end if

      !---------------- begin iterations -----------
      do iter = 1,niter+1

         if (iter==1) then
            if (.not. L_need_find_qnt_d) then
               call norm_init(rd_norm,rt_d,str_norm_dat)
            end if
            if (.not. L_need_find_qnt_m) then
               call norm_init(rm_norm,rt_m,str_norm_mod)
            end if
         end if

         !For search quantile case, re-search the quantile at iter=1,2,2+qt_period...
         if ( (iter==1 .or. mod(iter-2,qt_period)==0) .and.&
              &(L_need_find_qnt_d .or. L_need_find_qnt_m)) then
            if (L_need_find_qnt_d) then
               rt1_d=maxval(abs(rd))
               rt2_d=quantile( int(quant_d*size(rd)), abs(rd));
               rt_d = rt2_d

               write(0,*) "iter=",iter,"rt_d=",rt2_d,"max=",rt1_d
               !Init the Norm object for data residue
               call norm_init(rd_norm,rt_d,str_norm_dat)
            endif
            if (L_need_find_qnt_m) then
               rt1_m=maxval(abs(rm))
               rt2_m=quantile( int(quant_m*size(rm)), abs(rm));
               rt_m=rt2_m

               if (iter==1 .and. rt_m < 1e-6) then !!! rm is all zero at init
                  rt_m = 1e6
                  !set rt_m to a big value, and compute rt_m at next iteration
               end if
               write(0,*) "iter=",iter,"rt_m=",rt2_m,"max=",rt1_m               
               !Init the Norm object for model residue
               call norm_init(rm_norm,rt_m,str_norm_mod)
            end if
         end if

         !!-----------------Conjugate direction Algorithm-------------------
         ! Compute C'*Rd ==> g  = (WF)'*C'*Rd
         !         C'*Rm ==> g +=    A'*C'*Rm
         stat=norm_c1(rd_norm,rd,rwd)

         if (present( Wop)) then
            call chain0(Wop,Fop,T,F,g,rwd,td)
         else
            stat =          Fop(T,F,g,rwd   )  ! g  = (WF)'Rwd
         end if

         stat=norm_c1(rm_norm,rm,rwm)
         stat =  Aop(T,T,g,eps*rwm)                           
         !g +=  e A'Rm

         if (present( Jop)) then
            allocate(g_tmp(size(g)))
            g_tmp=g
            stat=Jop(F,F,g_tmp,g   )
         end if
         !g  =   J  g

         if (present( Wop)) then
            call chain0(Wop,Fop,F,F,g,gd,td)
         else
            stat =           Fop(F,F,g,gd   )
         end if
         !Gd  = (WF) g

         stat =  Aop(F,F,eps*g,gm)                           
         !Gm  =  e A g



         !compute the error statistics: 
         if (iter==1) then
            stat = norm_modulus(rd_norm,rd,sum_norm_dat0);
            ! write(0,*) "sum_norm_dat0 = ",sum_norm_dat0
            stat = norm_modulus(rm_norm,rm,sum_norm_mod0);
            ! write(0,*) "sum_norm_mod0 = ",sum_norm_mod0
            sum_norm0 = sum_norm_dat0 + sum_norm_mod0

            stat = compute_avg_r_from_c(rd_norm,sum_norm_dat0/size(rd),avg_rd0)
            ! write(0,*) "the inital avg_rd0 is", avg_rd0 
            stat = norm_modulus(g_norm,g,R_tmp)  ! L2 Norm, gd is a pointer!! length of gd might be misunderstood
            meas_g0 = sqrt(R_tmp)
         else
            stat = norm_modulus(g_norm,g,R_tmp)
            meas_g(iter-1) = sqrt(R_tmp)
            ! write(0,*) "iter=",iter, "the modulus of g =", meas_g(iter-1)
         end if

         !write(0,*) "iter=,m=",iter, m

         if (iter==niter+1) then
            exit  !exit loop
         end if

         ! Stepper run
         stat = stepper(forget, m,g, rr,gg)                  
         !m+=dm; R+=dR


         !Only stat eq 0 means normal
         if (stat .eq.1) then
            ! got stuck descending
            write (0,*) "Stuck in descending ==> Exit !"
            exit
         else if (stat.ne.0) then
            write(0,*) "stepper cease in advance"
            exit
                  end if
                

         if (exist_file("mod_iter")) then
            call sep_write(m,"mod_iter")
         end if
         ! ----------- report -----------
         if (present( mmov)) then
            mmov(:,iter) =  m(:size(mmov,1))
         end if
         if (present( rmov)) then
            rmov(:,iter) = rr(:size(rmov,1))
         end if
         if (present( err )) then
            err(  iter) = dot_product(rd,rd)
         end if
         if (present( verb)) then
            if (verb) then
               call solver_report(iter,m,g,rd,rm)
            end if
         end if
         forget=F

         ! -----------Extract error output-------
         !psiter has to be constant, otherwise this might be problematic
         ptr  => sum_norm_dat((iter-1)*psiter0+1 : iter*psiter0)
         ptr1 => sum_norm_mod((iter-1)*psiter0+1 : iter*psiter0)
         ptr3 => sum_norm((iter-1)*psiter0+1 : iter*psiter0)

         ptr2 => avg_rd((iter-1)*psiter0+1 : iter*psiter0)
         ptr4 => avg_rm((iter-1)*psiter0+1 : iter*psiter0)

         write(0,*) 'iter=', iter
         call ps_QueryError(err_out=ptr3,err_rd_out=ptr,avg_rd_out=ptr2,err_rm_out=ptr1, avg_rm_out=ptr4)
      end do

      if (present( resd)) then
         resd = rd
      end if
      if (present( resm)) then
         resm = rm(:size(resm))
      end if

      !close the plane-search module
      call ps_close()

    end subroutine solver_reg_nrm

 !-----------Function, Get all error statistics
  ! Output :
        !sum_norm_r0,sum_norm_r: The value of Objective Function
	!|r|, |d|
	!|F'C'(r), F'C'(d)|
       
  ! Return 0 means Success	
    integer function reg_nrm_get_error_stat(OF0,OF,OF_dat_res0,OF_dat_res,OF_mod_res0,OF_mod_res,d,r,Fd,Fr) 
   
      real, intent(out) ::OF0,d,Fd,OF_dat_res0,OF_mod_res0
      real, dimension(:), intent(out) ::OF,r,Fr,OF_dat_res,OF_mod_res  ! Array type
      
      reg_nrm_get_error_stat = 1

      OF0 = sum_norm0
      OF  = sum_norm
      OF_dat_res0 = sum_norm_dat0
      OF_dat_res  = sum_norm_dat
      OF_mod_res0 = sum_norm_mod0
      OF_mod_res = sum_norm_mod
      
      d = avg_rd0
      r = avg_rd

      Fd = meas_g0
      Fr = meas_g
      
      reg_nrm_get_error_stat = 0
    end function reg_nrm_get_error_stat
  
    subroutine solver_reg_nrm_close()
      deallocate(avg_rd,avg_rm,meas_g)
      deallocate(sum_norm,sum_norm_dat,sum_norm_mod)
    end subroutine solver_reg_nrm_close

end module 
