module solver_smp_nrm_mod
! 0 = W (F J m - d)

  use sep
  use chain0_mod 
  use solver_report_mod
  use norm_mod
  use plane_search_mod

  implicit none
  logical, parameter, private  :: T = .true., F = .false.
  
  !Store error information
  real,private		::avg_r0,meas_g0,sum_norm_r0 ! |r| and |F'C'(r)| intial value
  real,dimension(:),allocatable,target,private :: avg_r, sum_norm_r
  real,dimension(:),allocatable,private :: meas_g ! the norm of F'C'(r) ,using L2 currently
  
  contains
    subroutine solver_smp_nrm( m,d, Fop, stepper, niter , &
         &    norm_in,qnt,R_th_in,qt_period_in,psiter,overshot_shrink_max_cnt_in,Wop,Jop,m0,err,resd,mmov,rmov,verb)
      optional :: norm_in,qnt,R_th_in,qt_period_in,psiter,overshot_shrink_max_cnt_in,Wop,Jop,m0,err,resd,mmov,rmov,verb
      interface
         !-------------------------- begin definitions -----------

         integer function Fop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Fop
         integer function Wop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Wop
         integer function Jop(adj,add,m,d)
           real::m(:),d(:)
           logical,intent(in)::adj,add
         end function Jop
         integer function stepper(forget,m,g,rr,gg)
           real, dimension(:)  ::        m,g,rr,gg
           logical             :: forget
         end function stepper
      end interface
      real, dimension(:),    intent(in)          :: d, m0
      integer,               intent(in)          :: niter, psiter ! Iter. for solver and plane search

      integer,intent(in):: qt_period_in !Quantile Update mode  C or V  Constant/Variable
      integer:: qt_period

      logical,               intent(in)          :: verb
      character(12), intent(in)                           :: norm_in          ! Norm type to be used
      character(12) ::norm
      character(12) ::str_l2
      real,                  intent(in)          :: qnt           ! Quantile to set R_thr (R_bar)
      real :: R_th_in

      real, dimension(:),    intent(out)         :: m,err, resd
      real, dimension(:,:),  intent(out)         ::        rmov, mmov
      real, dimension(size(m))                   :: g
      real, dimension(size(d)),         target   :: rr, gg
      real, dimension(size(d)+size(m)), target   :: tt
      real, dimension(size(d))                   :: rwd
      real, dimension(:), pointer                :: rd, gd, td
      real, dimension(:), pointer                :: rm, gm, tm

      integer                                    :: iter, stat, psiter0
      logical                                    :: forget
      real                                       :: rt,rt1,rt2,quant    ! R_thr (R_bar), Quantile
      type(norm_obj) :: rd_norm   !norm object for rd
      type(norm_obj) :: rm_norm   !Dummy object here
      type(norm_obj) :: g_norm    !L2 norm for computing the gradient, measuring success
      logical L_need_find_qnt
      logical L_need_Rth

      integer        :: overshot_shrink_max_cnt_in
      real           :: R_tmp
      real, dimension(:), pointer ::ptr
      real, dimension(:), pointer ::ptr2


      rd => rr(1:size(d))
      gd => gg(1:size(d))
      td => tt(1:size(d))
      tm => tt(1+size(d):)

      ! -------- begin initialization --------
      psiter0=1
      if (present(psiter)) psiter0=psiter

      if (present( Wop)) then
         stat=Wop(F,F,-d,rd)
      else
         rd = -d
         !Rd = -W  d
      end if

      if (present( m0)) then
         m=m0
         !m  =     m0
         if (present( Wop)) then
            call chain0(Wop,Fop,F,T,m,rd,td)
         else
            stat =           Fop(F,T,m,rd   )
         end if
         !Rd+=  WF m0

      else
         m=0
      end if
      forget = T

      write(0,*) "The dim of m =", size(m), "the dim of rd = ", size(rd) 
      ! ------------- Set Norm, Quantile and Rt -------------
      if (.not. present(norm_in)) then
         norm="hybrid"
      else
         norm = norm_in
      end if

      quant=1.00
      L_need_find_qnt = .false.
      L_need_Rth = norm_need_Rth(norm)
      if (L_need_Rth) then
         if (present(R_th_in)) then
            rt = R_th_in
            L_need_find_qnt = .false.
         else
            L_need_find_qnt = .true.
            if (present(qnt)) then
               quant = qnt
            else
               call erexit("Neither quantile or R_th is specified")
            end if
         end if
      end if
      
      !Initalization error recording array
      allocate(avg_r(psiter0*niter),sum_norm_r(psiter0*niter))	
      allocate(meas_g(niter))

      write(0,*) "init g_norm l2 norm"
      str_l2 = "l2"
      call norm_init(g_norm,0.0,str_l2)

      qt_period = 10000 !Default don't search Rth for multiple times
      if (.not. present(qt_period_in)) then
         write(0,*) "qt_change_period not specified"
      else
         qt_period = qt_period_in
      end if


      !Init the plane_search_module
      !ps_init(psiter_in, ps_shrinkthres_in, len_datres_in, len_modres_in, len_mod, norm_datres_in, norm_modres_in)
      if (.not.present(overshot_shrink_max_cnt_in)) then
         call ps_init(psiter0,0,size(rd),  0,             size(m), rd_norm,        rm_norm,verb=.false.)
      else 
         call ps_init(psiter0,overshot_shrink_max_cnt_in,size(rd),  0,             size(m), rd_norm,        rm_norm, verb=.false.)
      end if

      !------------- begin iterations ------------
      do iter = 1,niter+1

         !Change Rthr through Quantine during iterations
         !if (iter==1 .or. iter==2 .or. MOD(iter-1,qt_period)==0) then
         if (iter==1) then
            call norm_init(rm_norm,1.0,norm) !Dumb
            if (.not. L_need_find_qnt) then
               call norm_init(rd_norm,rt,norm)
            end if
         end if

         !For search quantile case, re-search the quantile at iter=1,2,2+qt_period...
         if ((iter==1 .or. mod(iter-2,qt_period)==0) .and. L_need_find_qnt) then
            
            rt1=maxval(abs(rd));  !This will pick up biggest
           
            rt2=quantile( int(quant*size(rd)), abs(rd)); 
            !This will pick up the quant*size(rd)th smallest value in rd array
            rt=rt2;

				!If rt is too small, then use the maximum value in rd
				if (rt<1e-4*rt1) rt=rt1

				if (rt<1e-6) rt=1.0

            write(0,*) "iter=",iter,"quantile value=",rt2,"max=",rt1,"Init Norm object"
            !Init the Norm object

            call norm_init(rd_norm,rt,norm)
         end if



         !compute the first derivative of residue rwd, so then gradient delta(m) = F'(rwd)
         stat=norm_c1(rd_norm,rd,rwd)
         ! Compute C'*Rd ==> g= (WF)'*C'*Rd

         if (present(Wop)) then
            call chain0(Wop,Fop,T,F,g,rwd,td)
         else
            stat =          Fop(T,F,g,rwd)
            !g  = (WF)'Rwd
         end if

         if (present(Jop)) then
            tm=g
            stat  = Jop(F,F,tm, g  )
         end if
         !g  =   J  g

         if (present(Wop)) then
            call chain0(Wop,Fop,F,F,g,gd,td)
         else
            stat =          Fop(F,F,g,gd   )
            !Gd = (WF)' g 
         end if


         !compute the error statistics: 
         if (iter==1) then
            stat = norm_modulus(rd_norm,rd,sum_norm_r0)
            !write(0,*) "sum_norm_r0 = ",sum_norm_r0
            stat = compute_avg_r_from_c(rd_norm,sum_norm_r0/size(rd),avg_r0)
            !write(0,*) "the inital error c_r is ", 
            !write(0,*) "the inital avg_r0 is", avg_r0 
            stat = norm_modulus(g_norm,g,R_tmp)  ! L2 Norm, gd is a pointer!! length of gd might be misunderstood
            meas_g0 = sqrt(R_tmp)
         else
            stat = norm_modulus(g_norm,g,R_tmp)
            meas_g(iter-1) = sqrt(R_tmp)
            !write(0,*) "iter=",iter, "the modulus of g =", meas_g(iter-1)
         end if


         if (iter==niter+1) then
            exit  !exit loop
         end if

         ! Stepper run
         stat = stepper(forget, m,g, rr,gg)
         !m+=dm; R+=dR


         if (stat .eq.1) then
            ! got stuck descending
            write (0,*) "Stuck in descending ==> Exit !"
            exit
         end if

         if (exist_file("mod_iter")) then
            call sep_write(m,"mod_iter")
         end if

         ! ----------- report -----------
         if (present( mmov)) then
            mmov(:,iter) =  m(:size(mmov,1))
         end if
         if (present( rmov)) then
            rmov(:,iter) = rd(:size(rmov,1))
         end if
         if (present( err )) then
            err(  iter) = dot_product(rd,rd)
         end if
         if (present( verb)) then
            if (verb) then
               call solver_report(iter,m,g,rd)
            end if
         end if
         forget=F

         ! -----------Extract error output-------
         !psiter has to be constant, otherwise this might be problematic
	 ptr => sum_norm_r((iter-1)*psiter0+1 : iter*psiter0)
         ptr2 => avg_r((iter-1)*psiter0+1 : iter*psiter0)
         call ps_QueryError(ptr,ptr,ptr2)
      end do

      if (present( resd)) then
         resd = rd
      end if

      call ps_close()

    end subroutine solver_smp_nrm

  !-----------Function, Get all error statistics
  ! Output :
        !sum_norm_r0,sum_norm_r: The value of Objective Function
	!|r|, |d|
	!|F'C'(r), F'C'(d)|
       
  ! Return 0 means Success	
    integer function smp_nrm_get_error_stat(OF0,OF,d,r,Fd,Fr) 
   
      real, intent(out) ::OF0,d,Fd
      real, dimension(:), intent(out) ::OF,r,Fr  ! Array type
      
      smp_nrm_get_error_stat = 1

      OF0 = sum_norm_r0
      OF  = sum_norm_r

      d = avg_r0
      r = avg_r

      Fd = meas_g0
      Fr = meas_g

      smp_nrm_get_error_stat = 0
    end function smp_nrm_get_error_stat
  
  subroutine solver_smp_nrm_close()
	deallocate(sum_norm_r,avg_r,meas_g)
  end subroutine
end module
