module t_conv2_mod

  ! Transient Convolution, except that the length of xx equals to that of yy. 1-D
  use adj_mod

  implicit none

  integer, private :: n1_,n2_   !convolve the filter b on each trace
  real, dimension(:), pointer, private :: filter_, xx

contains
  !operator F, model is xx, matrix is filter, data is yy 
  subroutine t_conv2_init(filter_in,n1)
    real, dimension(:),target :: filter_in
    integer n1

    n1_ = n1
    n2_ = 1
    filter_  => filter_in

  end subroutine t_conv2_init


  function t_conv2_lop( adj, add, xx, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy 
    
    !If filter is longer than the data
    if (size(filter_) >= n1_) then
       write(0,*) size(filter_), n1_
       call erexit('the filter size is too large in conv2_lop')
    endif

    call adjnull (adj,add,xx,yy )
    call t_conv2_lop2(adj,add,xx,yy)
    stat=0
  end function t_conv2_lop

   subroutine t_conv2_lop2(adj,add,xx,yy)
    logical,intent(in) :: adj,add 
    real,dimension(:),target  :: xx,yy 
    integer :: i, stat
    real,  dimension (:), pointer  :: x_in,y_out

    !split the data into traces and convolve
    call tcad1_init2(filter_)
    if (adj) then
       do i=1,n2_
          x_in => xx((i-1)*n1_+1:i*n1_)
          y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcad1_lop2(adj,add,x_in,y_out)
       end do
    else
       do i=1,n2_
          x_in => xx((i-1)*n1_+1:i*n1_)
          y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcad1_lop2(adj,add,x_in,y_out)
       end do
    end if
  end subroutine t_conv2_lop2

  subroutine tcad1_init2 ( xx_in )
    real ,   dimension (:), pointer :: xx_in 
    xx => xx_in
  end subroutine tcad1_init2

  function tcad1_lop2 ( adj, add, bb, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy 
    call adjnull (adj,add,bb,yy )
    call tcad1_lop22(adj,add,bb,yy )
    stat=0
  end function tcad1_lop2

  subroutine tcad1_lop22(adj,add,bb,yy)
    logical,intent(in) :: adj,add 
    real, dimension (:)  :: bb 
    real, dimension (:)  :: yy 
    integer        x,   b,   y

    if ( size(yy) .ne. n1_ ) then
       call erexit('tcad1')
    end if
    
    !write(0,*) "max model, yy", maxval(bb), maxval(yy)
    !array xx is the filter, bb is model, yy is data
    do x= 1, size(xx)
       do b= 1, n1_ - x + 1
          y = x + b - 1
          if ( adj) then
             bb(b) =        bb(b) +  yy(y) * xx(x)
          else
             yy(y) =        yy(y) +  bb(b) * xx(x)
          end if
       end do
    end do
  end subroutine tcad1_lop22

end module t_conv2_mod

