module t_conv_filter
  ! Transient Convolution, except that the length of xx equals to that of yy. Adjoint is the Filter, 1-D
  use adj_mod

  implicit none

  integer, private :: n1_,n2_   !convolve the filter b on each trace
  real, dimension(:), pointer, private :: x_, xArr
  integer, private :: zero_lag_ !if zero_lag=0, then causual filter, otherwise >0, filter is double side

contains
  
  ! operator D, model is filter, matrix is data, data is yy
  subroutine t_conv_filter_init(x_in,n1,n2, zero_lag_in)
    real, dimension(:),target :: x_in
    integer n1,n2
    integer zero_lag_in
    optional :: zero_lag_in

    n1_ = n1
    n2_ = n2
    x_  => x_in

    if (present(zero_lag_in)) then
       zero_lag_ = zero_lag_in
    else
       zero_lag_ = 0
    end if
    if (zero_lag_ > 0) then
       write(0,*) "filter is non causal, lag=", zero_lag_in
    end if
    !write(0,*)  "x_",x_(1)
  end subroutine  t_conv_filter_init


  function t_conv_filter_lop( adj, add, bb, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy 

    !write(0,*) "enter t_conv_filter_lop"
 
    !write(0,*) "enter filter_lop, adj=", adj, size(bb), size(yy)   
    !If filter is longer than the data
    if (size(bb) >= n1_) then
       write(0,*) size(bb), n1_
       call erexit('the filter size is too large in conv_filter_lop')
    endif
 
    call adjnull (adj,add,bb,yy )
    call t_conv_filter_lop2(adj,add,bb,yy)
    stat=0
  end function t_conv_filter_lop

  function t_conv_filter_pad0_lop( adj, add, bb, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy
    real,dimension(:),allocatable  :: filter
    integer            :: szbb

    !Pad a zero at the beginning of the filter

    szbb = size(bb)
    if (szbb+1 >= n1_) then
       write(0,*) szbb, n1_
       call erexit('the filter size is too large in pad0_lap')
    endif


    allocate(filter(szbb+1))
    filter = 0

    filter(1+zero_lag_) = 0.0
    if (zero_lag_ > 0) then
       filter(1:zero_lag_) = bb(1:zero_lag_)
    end if

    if (zero_lag_ < szbb) then
       filter(zero_lag_+2:(szbb+1)) = bb(zero_lag_+1:szbb)
    end if

    call adjnull (adj,add,filter,yy)
    call t_conv_filter_lop2(adj,add,filter,yy)

    if (adj) then
       if (zero_lag_ > 0) then
          bb(1:zero_lag_) = filter(1:zero_lag_)
       end if
       if (zero_lag_ < szbb) then
          bb(zero_lag_+1:szbb)=filter(zero_lag_+2:(szbb+1))
       end if
    end if

    deallocate(filter)
    stat=0
  end function t_conv_filter_pad0_lop

   subroutine t_conv_filter_lop2(adj,add,bb,yy)
    logical,intent(in) :: adj,add 
    real,dimension(:),target  :: bb,yy 
    integer :: i, stat
    real,  dimension (:), pointer  :: x_in,y_out

    !split the data into traces and convolve
    if (adj) then
		 !$OMP PARALLEL DO SCHEDULE(STATIC) DEFAULT(SHARED) PRIVATE(i,x_in,y_out,stat)
       do i=1,n2_
          x_in => x_((i-1)*n1_+1:i*n1_)
			 y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcaf1_lop(adj,.true.,bb,y_out,x_in)
       end do
		 !$OMP END PARALLEL DO
    else
		 !$OMP PARALLEL DO SCHEDULE(STATIC) DEFAULT(SHARED) PRIVATE(i,x_in,y_out,stat)
       do i=1,n2_
          x_in => x_((i-1)*n1_+1:i*n1_)
          y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcaf1_lop(adj,add,bb,y_out,x_in)
       end do
		 !$OMP END PARALLEL DO
    end if
  end subroutine t_conv_filter_lop2

  subroutine tcaf1_init ( xx_in )
    real ,   dimension (:), pointer :: xx_in 
    xArr => xx_in
    !size(xx) should equal to n1
  end subroutine tcaf1_init

  function tcaf1_lop ( adj, add, bb, yy,xx) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy,xx 
    call adjnull (adj,add,bb,yy )
    call tcaf1_lop2(adj,add,bb,yy,xx)
    stat=0
  end function tcaf1_lop

  subroutine tcaf1_lop2(adj,add,bb,yy,xx)
    logical,intent(in) :: adj,add 
    real, dimension (:)  :: bb 
    real, dimension (:)  :: yy
	 real, dimension (:)  :: xx 
    integer        x,   b,   y
    !if ( size(yy) < size(xx) + size(bb) - 1 ) then
    !Here we change a little bit
    

    if ( size(yy) .ne. n1_ ) then
       call erexit('tcaf1')
    end if

    
    do b= 1, size(bb)
       do x= 1, n1_ - b + 1
          y = x + b - 1
    !write(0,*) "y,x,b", y,x,b
    !write(0,*)bb(b)
    !write(0,*) yy(y)
    !write(0,*) xx(x)
      
          if ( adj) then
             bb(b) =        bb(b) +  yy(y) * xx(x)
          else
             yy(y) =        yy(y) +  bb(b) * xx(x)
          end if
       end do
    end do
  end subroutine tcaf1_lop2

!!$  subroutine tcaf1_close()
!!$  end subroutine 
end module t_conv_filter
