module t_conv_filter_broke
  ! Transient Convolution, except that the length of xx equals to that of yy. Adjoint is the Filter, 1-D
  use adj_mod
  implicit none

  integer, private :: sz_
  integer, private :: n1_,n2_  !convolve the filter b on each trace
  real, dimension(:), pointer, private :: x_, xx
contains

  subroutine t_conv_filter_init(x_in,n1,n2)
    real, dimension(:) :: x_in
    integer n1,n2

    n1_ = n1
    n2_ = n2
    x_ = x_in
  end subroutine  t_conv_filter_init

  function t_conv_filter_lop( adj, add, bb, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy 

    call adjnull (adj,add,bb,yy )
    call t_conv_filter_lop2(adj,add,bb,yy)
    stat=0
  end function t_conv_filter_lop

   subroutine t_conv_filter_lop2(adj,add,bb,yy)
    logical,intent(in) :: adj,add 
    real,dimension(:),target  :: bb,yy 
    integer :: i, stat
    real,  dimension (:), pointer  :: x_in,y_out

    !split the data into traces and convolve
    if (adj) then
       do i=1,n2_
          x_in => x_((i-1)*n1_+1:i*n1_)
          call tcaf1_init(x_in)
          y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcaf1_lop(adj,.true.,bb,y_out)
       end do
    else
       do i=1,n2_
          x_in => x_((i-1)*n1_+1:i*n1_)
          call tcaf1_init(x_in)
          y_out=> yy((i-1)*n1_+1:i*n1_)
          stat = tcaf1_lop(adj,add,bb,y_out)
       end do
    end if
  end subroutine t_conv_filter_lop2

  subroutine tcaf1_init ( xx_in )
    real ,   dimension (:), pointer :: xx_in 
    xx => xx_in
    sz_ = size(xx)
    !sz_ should equal to n1
  end subroutine tcaf1_init

  function tcaf1_lop ( adj, add, bb, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: bb,yy 
    call adjnull (adj,add,bb,yy )
    call tcaf1_lop2(adj,add,bb,yy )
    stat=0
  end function tcaf1_lop

  subroutine tcaf1_lop2(adj,add,bb,yy)
    logical,intent(in) :: adj,add 
    real, dimension (:)  :: bb 
    real, dimension (:)  :: yy 
    integer        x,   b,   y
    !if ( size(yy) < size(xx) + size(bb) - 1 ) then
    !Here we change a little bit

    if ( size(yy) .ne. sz_ ) then
       call erexit('tcaf')
    end if


    do b= 1, size(bb)
       do x= 1, sz_ - b + 1
          y = x + b - 1
          if ( adj) then
             bb(b) =        bb(b) +  yy(y) * xx(x)
          else
             yy(y) =        yy(y) +  bb(b) * xx(x)
          end if
       end do
    end do
  end subroutine tcaf1_lop2

!!$  subroutine tcaf1_close()
!!$  end subroutine 
end module t_conv_filter_broke
